"""Warning checker."""
from typing import Dict, List, Sequence
from warnings import warn

from .spy_calls import BaseSpyCall, SpyCall, WhenRehearsal, VerifyRehearsal, match_call
from .warnings import MiscalledStubWarning, RedundantVerifyWarning


class WarningChecker:
    """An interface to inspect the call list and trigger any necessary warnings."""

    @staticmethod
    def check(all_calls: Sequence[BaseSpyCall]) -> None:
        """Check full call and rehearsal list for any potential misuse."""
        _check_no_miscalled_stubs(all_calls)
        _check_no_redundant_verify(all_calls)


def _check_no_miscalled_stubs(all_calls: Sequence[BaseSpyCall]) -> None:
    """Ensure every call matches a rehearsal, if the spy has rehearsals."""
    all_calls_by_id: Dict[int, List[BaseSpyCall]] = {}

    for call in all_calls:
        spy_id = call.spy_id
        spy_calls = all_calls_by_id.get(spy_id, [])
        all_calls_by_id[spy_id] = spy_calls + [call]

    for spy_id, spy_calls in all_calls_by_id.items():
        unmatched: List[SpyCall] = []

        for index, call in enumerate(spy_calls):
            past_stubs = [
                wr for wr in spy_calls[0:index] if isinstance(wr, WhenRehearsal)
            ]

            matched_past_stubs = [wr for wr in past_stubs if match_call(call, wr)]

            matched_future_verifies = [
                vr
                for vr in spy_calls[index + 1 :]
                if isinstance(vr, VerifyRehearsal) and match_call(call, vr)
            ]

            if (
                isinstance(call, SpyCall)
                and len(past_stubs) > 0
                and len(matched_past_stubs) == 0
                and len(matched_future_verifies) == 0
            ):
                unmatched = unmatched + [call]
                if index == len(spy_calls) - 1:
                    warn(MiscalledStubWarning(calls=unmatched, rehearsals=past_stubs))
            elif isinstance(call, WhenRehearsal) and len(unmatched) > 0:
                warn(MiscalledStubWarning(calls=unmatched, rehearsals=past_stubs))
                unmatched = []


def _check_no_redundant_verify(all_calls: Sequence[BaseSpyCall]) -> None:
    when_rehearsals = [c for c in all_calls if isinstance(c, WhenRehearsal)]
    verify_rehearsals = [c for c in all_calls if isinstance(c, VerifyRehearsal)]

    for vr in verify_rehearsals:
        if any(wr for wr in when_rehearsals if wr == vr):
            warn(RedundantVerifyWarning(rehearsal=vr))



"""Spy call handling."""
from typing import Any

from .call_stack import CallStack
from .stub_store import StubStore
from .spy_calls import SpyCall


class CallHandler:
    """An interface to handle calls to spies."""

    def __init__(self, call_stack: CallStack, stub_store: StubStore) -> None:
        """Initialize the CallHandler with access to SpyCalls and Stubs."""
        self._call_stack = call_stack
        self._stub_store = stub_store

    def handle(self, call: SpyCall) -> Any:
        """Handle a Spy's call, triggering stub behavior if necessary."""
        behavior = self._stub_store.get_by_call(call)
        self._call_stack.push(call)

        if behavior.error:
            raise behavior.error

        if behavior.action:
            return behavior.action(*call.args, **call.kwargs)

        return behavior.return_value



"""Warnings raised by Decoy.

See the [warnings guide][] for more details.

[warnings guide]: ../usage/errors-and-warnings/#warnings
"""
import os
from typing import Sequence

from .spy_calls import SpyCall, WhenRehearsal, VerifyRehearsal
from .stringify import stringify_call, stringify_error_message, count


class DecoyWarning(UserWarning):
    """Base class for Decoy warnings."""

    pass


class MiscalledStubWarning(DecoyWarning):
    """A warning when a configured Stub is called with non-matching arguments.

    This warning is raised if a mock is both:

    - Configured as a stub with [`when`][decoy.Decoy.when]
    - Called with arguments that do not match any configured behaviors

    See the [MiscalledStubWarning guide] for more details.

    [MiscalledStubWarning guide]: ../usage/errors-and-warnings/#miscalledstubwarning

    Attributes:
        rehearsals: The mocks's configured rehearsals.
        calls: Actual calls to the mock.
    """

    rehearsals: Sequence[WhenRehearsal]
    calls: Sequence[SpyCall]

    def __init__(
        self,
        rehearsals: Sequence[WhenRehearsal],
        calls: Sequence[SpyCall],
    ) -> None:
        heading = os.linesep.join(
            [
                "Stub was called but no matching rehearsal found.",
                f"Found {count(len(rehearsals), 'rehearsal')}:",
            ]
        )

        message = stringify_error_message(
            heading=heading,
            rehearsals=rehearsals,
            calls=calls,
        )

        super().__init__(message)
        self.rehearsals = rehearsals
        self.calls = calls


class RedundantVerifyWarning(DecoyWarning):
    """A warning when a mock is redundantly checked with `verify`.

    A `verify` assertion is redundant if:

    - A given call is used as a [`when`][decoy.Decoy.when] rehearsal
    - That same call is later used in a [`verify`][decoy.Decoy.verify] check

    See the [RedundantVerifyWarning guide][] for more details.

    [RedundantVerifyWarning guide]: ../usage/errors-and-warnings/#redundantverifywarning
    """

    def __init__(self, rehearsal: VerifyRehearsal) -> None:
        message = os.linesep.join(
            [
                "The same rehearsal was used in both a `when` and a `verify`.",
                "This is redundant and probably a misuse of the mock.",
                f"\t{stringify_call(rehearsal)}",
                "See https://mike.cousins.io/decoy/usage/errors-and-warnings/#miscalledstubwarning",  # noqa: E501
            ]
        )
        super().__init__(message)
        self.rehearsal = rehearsal


class IncorrectCallWarning(DecoyWarning):
    """A warning raised if a Decoy mock with a spec is called incorrectly.

    If a call to a Decoy mock is incorrect according to `inspect.signature`,
    this warning will be raised.

    See the [IncorrectCallWarning guide][] for more details.

    [IncorrectCallWarning guide]: ../usage/errors-and-warnings/#incorrectcallwarning
    """



"""Spy creation and storage."""
from typing import List, Sequence

from .spy_calls import BaseSpyCall, SpyCall, WhenRehearsal, VerifyRehearsal
from .errors import MissingRehearsalError


class CallStack:
    """SpyCall stack for all Spies in the Decoy container."""

    def __init__(self) -> None:
        """Initialize a stack for all SpyCalls."""
        self._stack: List[BaseSpyCall] = []

    def push(self, spy_call: SpyCall) -> None:
        """Add a new spy call to the stack."""
        self._stack.append(spy_call)

    def consume_when_rehearsal(self, ignore_extra_args: bool) -> WhenRehearsal:
        """Consume the last call to a Spy as a `when` rehearsal.

        This marks a call as a rehearsal but does not remove it from the stack.
        """
        try:
            call = self._stack[-1]
        except IndexError:
            raise MissingRehearsalError()
        if not isinstance(call, SpyCall):
            raise MissingRehearsalError()

        rehearsal = WhenRehearsal(*call)._replace(ignore_extra_args=ignore_extra_args)
        self._stack[-1] = rehearsal
        return rehearsal

    def consume_verify_rehearsals(
        self,
        count: int,
        ignore_extra_args: bool,
    ) -> List[VerifyRehearsal]:
        """Consume the last `count` calls to Spies as rehearsals.

        This marks calls as rehearsals but does not remove them from the stack.
        """
        calls = self._stack[-count:]

        if len(calls) != count or not all(isinstance(call, SpyCall) for call in calls):
            raise MissingRehearsalError()

        rehearsals = [
            VerifyRehearsal(*call)._replace(ignore_extra_args=ignore_extra_args)
            for call in calls
        ]
        self._stack[-count:] = rehearsals
        return rehearsals

    def get_by_rehearsals(self, rehearsals: Sequence[VerifyRehearsal]) -> List[SpyCall]:
        """Get all non-rehearsal calls to the spies in the given rehearsals."""
        return [
            call
            for call in self._stack
            if isinstance(call, SpyCall)
            and any(rehearsal.spy_id == call.spy_id for rehearsal in rehearsals)
        ]

    def get_all(self) -> List[BaseSpyCall]:
        """Get a list of all calls and rehearsals made."""
        return list(self._stack)

    def clear(self) -> None:
        """Remove all stored calls."""
        self._stack.clear()



"""Spy objects.

Classes in this module are heavily inspired by the
[unittest.mock library](https://docs.python.org/3/library/unittest.mock.html).
"""
from inspect import getattr_static, isclass, iscoroutinefunction, isfunction, signature
from functools import partial
from warnings import warn
from typing import get_type_hints, Any, Callable, Dict, NamedTuple, Optional

from .spy_calls import SpyCall
from .warnings import IncorrectCallWarning

CallHandler = Callable[[SpyCall], Any]


class SpyConfig(NamedTuple):
    """Spy configuration options passed to create_spy."""

    handle_call: CallHandler
    spec: Optional[Any] = None
    name: Optional[str] = None
    module_name: Optional[str] = None
    is_async: bool = False


def _get_type_hints(obj: Any) -> Dict[str, Any]:
    """Get type hints for an object, if possible.

    The builtin `typing.get_type_hints` may fail at runtime,
    e.g. if a type is subscriptable according to mypy but not
    according to Python.
    """
    try:
        return get_type_hints(obj)
    except Exception:
        return {}


class BaseSpy:
    """Spy object base class.

    - Pretends to be another class, if another class is given as a spec
    - Lazily constructs child spies when an attribute is accessed
    """

    def __init__(
        self,
        handle_call: CallHandler,
        spec: Optional[Any] = None,
        name: Optional[str] = None,
        module_name: Optional[str] = None,
    ) -> None:
        """Initialize a BaseSpy from a call handler and an optional spec object."""
        self._spec = spec
        self._handle_call: CallHandler = handle_call
        self._spy_children: Dict[str, BaseSpy] = {}

        self._name = name
        self._module_name = module_name

        if name is None and spec is not None and hasattr(spec, "__name__"):
            self._name = spec.__name__

        if (
            module_name is None
            and spec is not None
            and hasattr(spec, "__module__")
            and not isinstance(spec, partial)
        ):
            self._module_name = spec.__module__

        # ensure spy can pass inspect.signature checks
        try:
            self.__signature__ = signature(spec)  # type: ignore[arg-type]
        except Exception:
            pass

    @property  # type: ignore[misc]
    def __class__(self) -> Any:
        """Ensure Spy can pass `instanceof` checks."""
        if isclass(self._spec):
            return self._spec

        return type(self)

    def _call(self, *args: Any, **kwargs: Any) -> Any:
        spy_id = id(self)
        spy_name = (
            self._name
            if self._name
            else f"{type(self).__module__}.{type(self).__qualname__}"
        )

        if hasattr(self, "__signature__"):
            try:
                bound_args = self.__signature__.bind(*args, **kwargs)
            except TypeError as e:
                # stacklevel: 3 ensures warning is linked to call location
                warn(IncorrectCallWarning(e), stacklevel=3)
            else:
                args = bound_args.args
                kwargs = bound_args.kwargs

        return self._handle_call(SpyCall(spy_id, spy_name, args, kwargs))

    def __repr__(self) -> str:
        """Get a helpful string representation of the spy."""
        name = self._name

        if name and self._spec:
            if self._module_name:
                name = f"{self._module_name}.{name}"
            return f"<Decoy mock of {name}>"
        elif name:
            return f'<Decoy mock "{name}">'
        else:
            return "<Decoy mock>"

    def __getattr__(self, name: str) -> Any:
        """Get a property of the spy.

        Lazily constructs child spies, basing them on type hints if available.
        """
        # do not attempt to mock magic methods
        if name.startswith("__") and name.endswith("__"):
            return super().__getattribute__(name)

        # return previously constructed (and cached) child spies
        if name in self._spy_children:
            return self._spy_children[name]

        child_spec = None
        child_is_async = False
        child_name = f"{self._name}.{name}" if self._name is not None else name

        if isclass(self._spec):
            child_hint = _get_type_hints(self._spec).get(name)
            child_spec = getattr_static(self._spec, name, child_hint)

        if isinstance(child_spec, property):
            child_spec = _get_type_hints(child_spec.fget).get("return")

        if isinstance(child_spec, staticmethod):
            child_spec = child_spec.__func__

        elif isclass(self._spec) and isfunction(child_spec):
            # `iscoroutinefunction` does not work for `partial` on Python < 3.8
            # check before we wrap it
            child_is_async = iscoroutinefunction(child_spec)

            # consume the `self` argument of the method to ensure proper
            # signature reporting by wrapping it in a partial
            child_spec = partial(child_spec, None)  # type: ignore[arg-type]

        spy = create_spy(
            config=SpyConfig(
                handle_call=self._handle_call,
                spec=child_spec,
                name=child_name,
                module_name=self._module_name,
                is_async=child_is_async,
            ),
        )

        self._spy_children[name] = spy

        return spy


class Spy(BaseSpy):
    """An object that records all calls made to itself and its children."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Handle a call to the spy."""
        return self._call(*args, **kwargs)


class AsyncSpy(BaseSpy):
    """An object that records all async. calls made to itself and its children."""

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Handle a call to the spy asynchronously."""
        return self._call(*args, **kwargs)


SpyFactory = Callable[[SpyConfig], Any]


def create_spy(config: SpyConfig) -> Any:
    """Create a Spy from a spec.

    Functions and classes passed to `spec` will be inspected (and have any type
    annotations inspected) to ensure `AsyncSpy`'s are returned where necessary.
    """
    handle_call, spec, name, module_name, is_async = config
    _SpyCls = AsyncSpy if iscoroutinefunction(spec) or is_async is True else Spy

    return _SpyCls(
        handle_call=handle_call,
        spec=spec,
        name=name,
        module_name=module_name,
    )



"""Spy call verification."""
from typing import Optional, Sequence

from .spy_calls import SpyCall, VerifyRehearsal, match_call
from .errors import VerifyError

# ensure decoy.verifier does not pollute Pytest tracebacks
__tracebackhide__ = True


class Verifier:
    """An interface to verify that spies were called as expected."""

    @staticmethod
    def verify(
        rehearsals: Sequence[VerifyRehearsal],
        calls: Sequence[SpyCall],
        times: Optional[int] = None,
    ) -> None:
        """Verify that a list of calls satisfies a given list of rehearsals.

        Arguments:
            rehearsals: Rehearsal calls to verify.
            calls: All calls made to the spies in the rehearsals list.
            times: Number of times the rehearsal sequence should appear in calls.
                If omitted, will look for "at least once."

        Raises:
            VerifyError: the actual calls to the spy did not match the given
                rehearsals.
        """
        match_count = 0

        for i in range(len(calls)):
            calls_subset = calls[i : i + len(rehearsals)]
            matches = [match_call(c, r) for c, r in zip(calls_subset, rehearsals)]

            if all(matches) and len(calls_subset) == len(rehearsals):
                match_count = match_count + 1

        calls_verified = match_count != 0 if times is None else match_count == times

        if not calls_verified:
            raise VerifyError(
                rehearsals=rehearsals,
                calls=calls,
                times=times,
            )



"""Spy call value objects."""
from typing import Any, Dict, NamedTuple, Tuple


class BaseSpyCall(NamedTuple):
    """A value object representing a call to a spy.

    This base class should not be used directly.

    Attributes:
        spy_id: Identifier of the spy that made the call.
        spy_name: String name of the spy.
        args: Arguments list of the call.
        kwargs: Keyword arguments list of the call.
        ignore_extra_args: Whether extra arguments to the call should
            be ignored during comparison. Only used by rehearsals.
    """

    spy_id: int
    spy_name: str
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    ignore_extra_args: bool = False


class BaseSpyRehearsal(BaseSpyCall):
    """A base class for rehearsals made to `when` or `verify`.

    This base class should not be used directly.
    """

    pass


class SpyCall(BaseSpyCall):
    """An call made to the spy by the code under test."""

    pass


class WhenRehearsal(BaseSpyRehearsal):
    """A call that has been used as a rehearsal for `when`."""

    pass


class VerifyRehearsal(BaseSpyRehearsal):
    """A call that has been used as a rehearsal for `when`."""

    pass


def match_call(call: BaseSpyCall, rehearsal: BaseSpyRehearsal) -> bool:
    """Check if a call matches a given rehearsal."""
    if call.spy_id != rehearsal.spy_id:
        return False

    if rehearsal.ignore_extra_args:
        try:
            args_match = all(
                call.args[i] == value for i, value in enumerate(rehearsal.args)
            )
            kwargs_match = all(
                call.kwargs[key] == value for key, value in rehearsal.kwargs.items()
            )

            return args_match and kwargs_match

        except (IndexError, KeyError):
            return False

    return call.args == rehearsal.args and call.kwargs == rehearsal.kwargs



"""Decoy mypy plugin."""
from typing import Callable, Optional, Type as ClassType

from mypy.errorcodes import FUNC_RETURNS_VALUE
from mypy.plugin import Plugin, MethodContext
from mypy.types import Type


class DecoyPlugin(Plugin):
    """A mypy plugin to remove otherwise valid errors from rehearsals."""

    def get_method_hook(
        self, fullname: str
    ) -> Optional[Callable[[MethodContext], Type]]:
        """Remove any func-returns-value errors inside `when` or `verify` calls."""
        if fullname in {"decoy.Decoy.verify", "decoy.Decoy.when"}:
            return self._handle_decoy_call

        return None

    def _handle_decoy_call(self, ctx: MethodContext) -> Type:
        errors_list = ctx.api.msg.errors.error_info_map.get(ctx.api.path, [])
        rehearsal_call_args = ctx.args[0] if len(ctx.args) > 0 else []
        error_removals = []

        for err in errors_list:
            if err.code == FUNC_RETURNS_VALUE:
                for arg in rehearsal_call_args:
                    if arg.line == err.line and arg.column == err.column:
                        error_removals.append(err)

        for err in error_removals:
            errors_list.remove(err)

        return ctx.default_return_type


def plugin(version: str) -> ClassType[DecoyPlugin]:
    """Get the DecoyPlugin class definition."""
    return DecoyPlugin



"""Common type definitions."""
from typing import Any, Callable, TypeVar

FuncT = TypeVar("FuncT", bound=Callable[..., Any])
"""A function to create a decoy of."""

ClassT = TypeVar("ClassT", bound=object)
"""A class to create a decoy of."""

ReturnT = TypeVar("ReturnT")
"""The return type of a given call."""



"""Decoy stubbing and spying library."""
from typing import Any, Callable, Generic, Optional, cast, overload

from . import matchers, errors, warnings
from .core import DecoyCore, StubCore
from .types import ClassT, FuncT, ReturnT

# ensure decoy does not pollute pytest tracebacks
__tracebackhide__ = True


class Decoy:
    """Decoy test double state container."""

    def __init__(self) -> None:
        """Initialize the state container for test doubles and stubs.

        You should initialize a new Decoy instance for every test. See the
        [setup guide](../#setup) for more details.
        """
        self._core = DecoyCore()

    @overload
    def mock(self, *, cls: Callable[..., ClassT]) -> ClassT:
        ...

    @overload
    def mock(self, *, func: FuncT) -> FuncT:
        ...

    # TODO(mc, 2021-11-14): make `name` required for specless mocks in v2.0
    @overload
    def mock(self, *, name: Optional[str] = None, is_async: bool = False) -> Any:
        ...

    def mock(
        self,
        *,
        cls: Optional[Any] = None,
        func: Optional[Any] = None,
        name: Optional[str] = None,
        is_async: bool = False,
    ) -> Any:
        """Create a mock.

        See the [mock creation guide](../usage/create/) for more details.

        Arguments:
            cls: A class definition that the mock should imitate.
            func: A function definition the mock should imitate.
            name: A name to use for the mock. If you do not use
                `cls` or `func`, you should add a `name`.
            is_async: Force the returned spy to be asynchronous. This argument
                only applies if you don't use `cls` nor `func`.

        Returns:
            A spy typecast as the object it's imitating, if any.

        Example:
            ```python
            def test_get_something(decoy: Decoy):
                db = decoy.mock(cls=Database)
                # ...
            ```
        """
        spec = cls or func
        return self._core.mock(spec=spec, name=name, is_async=is_async)

    def create_decoy(
        self,
        spec: Callable[..., ClassT],
        *,
        is_async: bool = False,
    ) -> ClassT:
        """Create a class mock for `spec`.

        !!! warning "Deprecated since v1.6.0"
            Use [decoy.Decoy.mock][] with the `cls` parameter, instead.
        """
        spy = self._core.mock(spec=spec, is_async=is_async)
        return cast(ClassT, spy)

    def create_decoy_func(
        self,
        spec: Optional[FuncT] = None,
        *,
        is_async: bool = False,
    ) -> FuncT:
        """Create a function mock for `spec`.

        !!! warning "Deprecated since v1.6.0"
            Use [decoy.Decoy.mock][] with the `func` parameter, instead.
        """
        spy = self._core.mock(spec=spec, is_async=is_async)
        return cast(FuncT, spy)

    def when(
        self,
        _rehearsal_result: ReturnT,
        *,
        ignore_extra_args: bool = False,
    ) -> "Stub[ReturnT]":
        """Create a [Stub][decoy.Stub] configuration using a rehearsal call.

        See [stubbing usage guide](../usage/when/) for more details.

        Arguments:
            _rehearsal_result: The return value of a rehearsal, used for typechecking.
            ignore_extra_args: Allow the rehearsal to specify fewer arguments than
                the actual call. Decoy will compare and match any given arguments,
                ignoring unspecified arguments.

        Returns:
            A stub to configure using `then_return`, `then_raise`, or `then_do`.

        Example:
            ```python
            db = decoy.mock(cls=Database)
            decoy.when(db.exists("some-id")).then_return(True)
            ```

        Note:
            The "rehearsal" is an actual call to the test fake. Because the
            call is written inside `when`, Decoy is able to infer that the call
            is a rehearsal for stub configuration purposes rather than a call
            from the code-under-test.
        """
        stub_core = self._core.when(
            _rehearsal_result,
            ignore_extra_args=ignore_extra_args,
        )
        return Stub(core=stub_core)

    def verify(
        self,
        *_rehearsal_results: Any,
        times: Optional[int] = None,
        ignore_extra_args: bool = False,
    ) -> None:
        """Verify a decoy was called using one or more rehearsals.

        See [verification usage guide](../usage/verify/) for more details.

        Arguments:
            _rehearsal_results: The return value of rehearsals, unused except
                to determine how many rehearsals to verify.
            times: How many times the call should appear. If `times` is specified,
                the call count must match exactly, otherwise the call must appear
                at least once. The `times` argument must be used with exactly one
                rehearsal.
            ignore_extra_args: Allow the rehearsal to specify fewer arguments than
                the actual call. Decoy will compare and match any given arguments,
                ignoring unspecified arguments.

        Example:
            ```python
            def test_create_something(decoy: Decoy):
                gen_id = decoy.mock(func=generate_unique_id)

                # ...

                decoy.verify(gen_id("model-prefix_"))
            ```

        Note:
            A "rehearsal" is an actual call to the test fake. The fact that
            the call is written inside `verify` is purely for typechecking and
            API sugar. Decoy will pop the last call(s) to _any_ fake off its
            call stack, which will end up being the call inside `verify`.
        """
        self._core.verify(
            *_rehearsal_results,
            times=times,
            ignore_extra_args=ignore_extra_args,
        )

    def reset(self) -> None:
        """Reset all decoy state.

        This method should be called after every test to ensure spies and stubs
        don't leak between tests. The Decoy fixture provided by the pytest plugin
        will do this automatically.

        The `reset` method may also trigger warnings if Decoy detects any questionable
        mock usage. See [decoy.warnings][] for more details.
        """
        self._core.reset()


class Stub(Generic[ReturnT]):
    """A rehearsed Stub that can be used to configure mock behaviors.

    See [stubbing usage guide](../usage/when/) for more details.
    """

    def __init__(self, core: StubCore) -> None:
        self._core = core

    def then_return(self, *values: ReturnT) -> None:
        """Configure the stub to return value(s).

        Arguments:
            *values: Zero or more return values. Multiple values will result
                     in different return values for subsequent calls, with the
                     last value latching in once all other values have returned.
        """
        self._core.then_return(*values)

    def then_raise(self, error: Exception) -> None:
        """Configure the stub to raise an error.

        Arguments:
            error: The error to raise.

        Note:
            Setting a stub to raise will prevent you from writing new
            rehearsals, because they will raise. If you need to make more calls
            to `when`, you'll need to wrap your rehearsal in a `try`.
        """
        self._core.then_raise(error)

    def then_do(self, action: Callable[..., ReturnT]) -> None:
        """Configure the stub to trigger an action.

        Arguments:
            action: The function to call. Called with whatever arguments
                are actually passed to the stub.
        """
        self._core.then_do(action)


__all__ = ["Decoy", "Stub", "matchers", "warnings", "errors"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['decoy', 'decoy.mypy']

package_data = \
{'': ['*']}

entry_points = \
{'pytest11': ['decoy = decoy.pytest_plugin']}

setup_kwargs = {
    'name': 'decoy',
    'version': '1.8.0',
    'description': 'Opinionated mocking library for Python',
    'long_description': '<div align="center">\n    <img alt="Decoy logo" src="https://mike.cousins.io/decoy/img/decoy.png" width="256px">\n    <h1 class="decoy-title">Decoy</h1>\n    <p>Opinionated mocking library for Python</p>\n    <p>\n        <a title="CI Status" href="https://github.com/mcous/decoy/actions">\n        <img src="https://img.shields.io/github/workflow/status/mcous/decoy/Continuous%20integration/main?style=flat-square"></a>\n        <a title="Code Coverage" href="https://app.codecov.io/gh/mcous/decoy/"><img src="https://img.shields.io/codecov/c/github/mcous/decoy?style=flat-square"></a>\n        <a title="License" href="https://github.com/mcous/decoy/blob/main/LICENSE"><img src="https://img.shields.io/github/license/mcous/decoy?style=flat-square"></a>\n        <a title="PyPI Version"href="https://pypi.org/project/decoy/"><img src="https://img.shields.io/pypi/v/decoy?style=flat-square"></a>\n        <a title="Supported Python Versions" href="https://pypi.org/project/decoy/"><img src="https://img.shields.io/pypi/pyversions/decoy?style=flat-square"></a>\n    </p>\n    <p>\n        <a href="https://mike.cousins.io/decoy/">Usage guide and documentation</a>\n    </p>\n</div>\n\nDecoy is a mocking library designed for **effective and productive test-driven development** in Python. If you want to use tests to guide the structure of your code, Decoy might be for you!\n\nDecoy mocks are **async/await** and **type-checking** friendly. Decoy is heavily inspired by (and/or stolen from) the excellent [testdouble.js][] and [Mockito][] projects. The Decoy API is powerful, easy to read, and strives to help you make good decisions about your code.\n\n## Install\n\n```bash\n# pip\npip install decoy\n\n# poetry\npoetry add --dev decoy\n\n# pipenv\npipenv install --dev decoy\n```\n\n## Setup\n\n### Pytest setup\n\nDecoy ships with its own [pytest][] plugin, so once Decoy is installed, you\'re ready to start using it via its pytest fixture, called `decoy`.\n\n```python\n# test_my_thing.py\nfrom decoy import Decoy\n\ndef test_my_thing_works(decoy: Decoy) -> None:\n    ...\n```\n\n### Mypy Setup\n\nBy default, Decoy is compatible with Python [typing][] and type-checkers like [mypy][]. However, stubbing functions that return `None` can trigger a [type checking error](https://mypy.readthedocs.io/en/stable/error_code_list.html#check-that-called-function-returns-a-value-func-returns-value) during correct usage of the Decoy API. To suppress these errors, add Decoy\'s plugin to your mypy configuration.\n\n```ini\n# mypy.ini\nplugins = decoy.mypy\n```\n\n## Basic Usage\n\nThis example assumes you are using [pytest][]. See Decoy\'s [documentation][] for a more detailed usage guide and API reference.\n\n### Define your test\n\nDecoy will add a `decoy` fixture that provides its mock creation API.\n\n```python\nfrom decoy import Decoy\nfrom todo import TodoAPI, TodoItem\nfrom todo.store TodoStore\n\ndef test_add_todo(decoy: Decoy) -> None:\n    ...\n```\n\n### Create a mock\n\nUse `decoy.mock` to create a mock based on some specification. From there, inject the mock into your test subject.\n\n```python\ndef test_add_todo(decoy: Decoy) -> None:\n    todo_store = decoy.mock(cls=TodoStore)\n    subject = TodoAPI(store=todo_store)\n    ...\n```\n\nSee [creating mocks][] for more details.\n\n### Stub a behavior\n\nUse `decoy.when` to configure your mock\'s behaviors. For example, you can set the mock to return a certain value when called in a certain way using `then_return`:\n\n```python\ndef test_add_todo(decoy: Decoy) -> None:\n    """Adding a todo should create a TodoItem in the TodoStore."""\n    todo_store = decoy.mock(cls=TodoStore)\n    subject = TodoAPI(store=todo_store)\n\n    decoy.when(\n        todo_store.add(name="Write a test for adding a todo")\n    ).then_return(\n        TodoItem(id="abc123", name="Write a test for adding a todo")\n    )\n\n    result = subject.add("Write a test for adding a todo")\n    assert result == TodoItem(id="abc123", name="Write a test for adding a todo")\n```\n\nSee [stubbing with when][] for more details.\n\n### Verify a call\n\nUse `decoy.verify` to assert that a mock was called in a certain way. This is best used with dependencies that are being used for their side-effects and don\'t return a useful value.\n\n```python\ndef test_remove_todo(decoy: Decoy) -> None:\n    """Removing a todo should remove the item from the TodoStore."""\n    todo_store = decoy.mock(cls=TodoStore)\n    subject = TodoAPI(store=todo_store)\n\n    subject.remove("abc123")\n\n    decoy.verify(todo_store.remove(id="abc123"))\n```\n\nSee [spying with verify][] for more details.\n\n[testdouble.js]: https://github.com/testdouble/testdouble.js\n[mockito]: https://site.mockito.org/\n[pytest]: https://docs.pytest.org/\n[typing]: https://docs.python.org/3/library/typing.html\n[mypy]: https://mypy.readthedocs.io/\n[documentation]: https://mike.cousins.io/decoy/\n[creating mocks]: https://mike.cousins.io/decoy/usage/create/\n[stubbing with when]: https://mike.cousins.io/decoy/usage/when/\n[spying with verify]: https://mike.cousins.io/decoy/usage/verify/\n',
    'author': 'Mike Cousins',
    'author_email': 'mike@cousins.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://mike.cousins.io/decoy/',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Decoy implementation logic."""
from typing import Any, Callable, Optional

from .spy import SpyConfig, SpyFactory, create_spy as default_create_spy
from .spy_calls import WhenRehearsal
from .call_stack import CallStack
from .stub_store import StubStore, StubBehavior
from .call_handler import CallHandler
from .verifier import Verifier
from .warning_checker import WarningChecker
from .types import ReturnT

# ensure decoy.core does not pollute Pytest tracebacks
__tracebackhide__ = True


class DecoyCore:
    """The DecoyCore class implements the main logic of Decoy."""

    def __init__(
        self,
        create_spy: Optional[SpyFactory] = None,
        verifier: Optional[Verifier] = None,
        warning_checker: Optional[WarningChecker] = None,
        stub_store: Optional[StubStore] = None,
        call_stack: Optional[CallStack] = None,
        call_handler: Optional[CallHandler] = None,
    ) -> None:
        """Initialize the DecoyCore with its dependencies."""
        self._create_spy = create_spy or default_create_spy
        self._verifier = verifier or Verifier()
        self._warning_checker = warning_checker or WarningChecker()
        self._stub_store = stub_store or StubStore()
        self._call_stack = call_stack or CallStack()
        self._call_hander = call_handler or CallHandler(
            call_stack=self._call_stack,
            stub_store=self._stub_store,
        )

    def mock(
        self,
        *,
        spec: Optional[Any] = None,
        name: Optional[str] = None,
        is_async: bool = False,
    ) -> Any:
        """Create and register a new spy."""
        config = SpyConfig(
            spec=spec,
            name=name,
            is_async=is_async,
            handle_call=self._call_hander.handle,
        )
        return self._create_spy(config)

    def when(self, _rehearsal: ReturnT, *, ignore_extra_args: bool) -> "StubCore":
        """Create a new stub from the last spy rehearsal."""
        rehearsal = self._call_stack.consume_when_rehearsal(
            ignore_extra_args=ignore_extra_args
        )
        return StubCore(rehearsal=rehearsal, stub_store=self._stub_store)

    def verify(
        self,
        *_rehearsals: ReturnT,
        times: Optional[int],
        ignore_extra_args: bool,
    ) -> None:
        """Verify that a Spy or Spies were called."""
        rehearsals = self._call_stack.consume_verify_rehearsals(
            count=len(_rehearsals),
            ignore_extra_args=ignore_extra_args,
        )
        calls = self._call_stack.get_by_rehearsals(rehearsals)

        self._verifier.verify(rehearsals=rehearsals, calls=calls, times=times)

    def reset(self) -> None:
        """Reset and remove all stored spies and stubs."""
        calls = self._call_stack.get_all()
        self._warning_checker.check(calls)
        self._call_stack.clear()
        self._stub_store.clear()


class StubCore:
    """The StubCore class implements the main logic of a Decoy Stub."""

    def __init__(self, rehearsal: WhenRehearsal, stub_store: StubStore) -> None:
        """Initialize the Stub with a configuration."""
        self._rehearsal = rehearsal
        self._stub_store = stub_store

    def then_return(self, *values: ReturnT) -> None:
        """Set the stub to return value(s)."""
        for i, return_value in reversed(list(enumerate(values))):
            self._stub_store.add(
                rehearsal=self._rehearsal,
                behavior=StubBehavior(
                    return_value=return_value,
                    once=(i != len(values) - 1),
                ),
            )

    def then_raise(self, error: Exception) -> None:
        """Set the stub to raise an error."""
        self._stub_store.add(
            rehearsal=self._rehearsal,
            behavior=StubBehavior(error=error),
        )

    def then_do(self, action: Callable[..., ReturnT]) -> None:
        """Set the stub to perform an action."""
        self._stub_store.add(
            rehearsal=self._rehearsal,
            behavior=StubBehavior(action=action),
        )
